# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# ruff: noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: multi_pex_api.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13multi_pex_api.proto\x12\rmulti_pex_api\"\x07\n\x05\x45mpty\"8\n\x16\x43reatePexServerRequest\x12\x1e\n\x16\x63reate_pex_server_args\x18\x01 \x01(\t\":\n\x14\x43reatePexServerReply\x12\"\n\x1a\x63reate_pex_server_response\x18\x01 \x01(\t\"4\n\x14GetPexServersRequest\x12\x1c\n\x14get_pex_servers_args\x18\x01 \x01(\t\"6\n\x12GetPexServersReply\x12 \n\x18get_pex_servers_response\x18\x01 \x01(\t\"C\n\x1bGetCrashedPexServersRequest\x12$\n\x1cget_crashed_pex_servers_args\x18\x01 \x01(\t\"E\n\x19GetCrashedPexServersReply\x12(\n get_crashed_pex_servers_response\x18\x01 \x01(\t\"<\n\x18ShutdownPexServerRequest\x12 \n\x18shutdown_pex_server_args\x18\x01 \x01(\t\">\n\x16ShutdownPexServerReply\x12$\n\x1cshutdown_pex_server_response\x18\x01 \x01(\t\"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\"\x19\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t2\xe0\x03\n\x0bMultiPexApi\x12_\n\x0f\x43reatePexServer\x12%.multi_pex_api.CreatePexServerRequest\x1a#.multi_pex_api.CreatePexServerReply\"\x00\x12Y\n\rGetPexServers\x12#.multi_pex_api.GetPexServersRequest\x1a!.multi_pex_api.GetPexServersReply\"\x00\x12n\n\x14GetCrashedPexServers\x12*.multi_pex_api.GetCrashedPexServersRequest\x1a(.multi_pex_api.GetCrashedPexServersReply\"\x00\x12\x65\n\x11ShutdownPexServer\x12\'.multi_pex_api.ShutdownPexServerRequest\x1a%.multi_pex_api.ShutdownPexServerReply\"\x00\x12>\n\x04Ping\x12\x1a.multi_pex_api.PingRequest\x1a\x18.multi_pex_api.PingReply\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'multi_pex_api_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EMPTY']._serialized_start=38
  _globals['_EMPTY']._serialized_end=45
  _globals['_CREATEPEXSERVERREQUEST']._serialized_start=47
  _globals['_CREATEPEXSERVERREQUEST']._serialized_end=103
  _globals['_CREATEPEXSERVERREPLY']._serialized_start=105
  _globals['_CREATEPEXSERVERREPLY']._serialized_end=163
  _globals['_GETPEXSERVERSREQUEST']._serialized_start=165
  _globals['_GETPEXSERVERSREQUEST']._serialized_end=217
  _globals['_GETPEXSERVERSREPLY']._serialized_start=219
  _globals['_GETPEXSERVERSREPLY']._serialized_end=273
  _globals['_GETCRASHEDPEXSERVERSREQUEST']._serialized_start=275
  _globals['_GETCRASHEDPEXSERVERSREQUEST']._serialized_end=342
  _globals['_GETCRASHEDPEXSERVERSREPLY']._serialized_start=344
  _globals['_GETCRASHEDPEXSERVERSREPLY']._serialized_end=413
  _globals['_SHUTDOWNPEXSERVERREQUEST']._serialized_start=415
  _globals['_SHUTDOWNPEXSERVERREQUEST']._serialized_end=475
  _globals['_SHUTDOWNPEXSERVERREPLY']._serialized_start=477
  _globals['_SHUTDOWNPEXSERVERREPLY']._serialized_end=539
  _globals['_PINGREQUEST']._serialized_start=541
  _globals['_PINGREQUEST']._serialized_end=568
  _globals['_PINGREPLY']._serialized_start=570
  _globals['_PINGREPLY']._serialized_end=595
  _globals['_MULTIPEXAPI']._serialized_start=598
  _globals['_MULTIPEXAPI']._serialized_end=1078
# @@protoc_insertion_point(module_scope)
